#include "config.h"

#include <gtk/gtk.h>
#include <xmms/configfile.h>
#include "goom_xmms.h"

static GtkWidget *configure_win = NULL;
static void adj_value_changed(GtkWidget *w, int *a);
static void ok_clicked(GtkWidget *w, GtkWidget *window);
static void cancel_clicked(GtkWidget *w, GtkWidget *window);
static void apply_clicked(GtkWidget *w, GtkWidget *window);
static gboolean check_cfg_version(ConfigFile *f);    
static void pixeldoubling_toggled(GtkWidget *w,int *data);
static void fullscreen_toggled(GtkWidget *w,int *data);
static void cinemascope_toggled(GtkWidget *w,int *data);

void plug_save_prefs(void) ;
void plug_load_prefs(void) ;

t_config config = {320,240,0,0} ;
t_config extern_config = {320,240,0,0} ;

static GtkWidget *fsButton = NULL; /* check button for fullscreen */

void plug_configure(void)
{
	GtkWidget *buttonbox,*ok,*cancel,*cbutton;
	static GtkWidget *vbox,*vbox2;
      
	GtkObject *xadj,*yadj,*sadj,*teadj,*tcadj;
	GtkWidget *label,*xhscale,*yhscale,*shscale;
	static GtkWidget *hbox,*frame;
	
	
	if(configure_win)
		return;

	plug_load_prefs();
	
        configure_win = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_widget_set_usize(configure_win, 558, 320);

        gtk_container_set_border_width(GTK_CONTAINER(configure_win), 10);
        gtk_window_set_title(GTK_WINDOW(configure_win), "What a GOOM! Configuration");
        gtk_window_set_policy(GTK_WINDOW(configure_win), FALSE, FALSE, FALSE);
        gtk_window_set_position(GTK_WINDOW(configure_win), GTK_WIN_POS_MOUSE);
	gtk_signal_connect(GTK_OBJECT(configure_win), "destroy", 
			   GTK_SIGNAL_FUNC(gtk_widget_destroyed), &configure_win);

	vbox = gtk_vbox_new(FALSE, 4);
	gtk_container_add(GTK_CONTAINER(configure_win),vbox);

	frame = gtk_frame_new("Display:");
	gtk_box_pack_start(GTK_BOX(vbox),frame,TRUE,TRUE,4);

	vbox2 = gtk_vbox_new(FALSE, 4);
	gtk_container_add(GTK_CONTAINER(frame),vbox2);
	
	hbox = gtk_hbox_new(FALSE, 4);

	label = gtk_label_new("X Resolution");
	gtk_box_pack_start(GTK_BOX(hbox),label,FALSE,FALSE,4);
	gtk_widget_show(label);

	xadj=gtk_adjustment_new(config.xres, 32, 1280, 32, 32, 32);
	xhscale=gtk_hscale_new(GTK_ADJUSTMENT(xadj));
	gtk_scale_set_draw_value(GTK_SCALE(xhscale), TRUE);
	gtk_scale_set_value_pos(GTK_SCALE(xhscale), GTK_POS_LEFT);
	gtk_scale_set_digits(GTK_SCALE(xhscale), 0);
	gtk_widget_show(xhscale);
	gtk_box_pack_start(GTK_BOX(hbox), xhscale, TRUE, TRUE, 5);

	gtk_box_pack_start(GTK_BOX(vbox2),hbox,TRUE,TRUE,4);
	gtk_widget_show(hbox);

	hbox = gtk_hbox_new(FALSE, 4);

	label = gtk_label_new("Y Resolution");
	gtk_box_pack_start(GTK_BOX(hbox),label,FALSE,FALSE,4);
	gtk_widget_show(label);

	yadj=gtk_adjustment_new(config.yres, 24, 960, 24, 24, 24);
	yhscale=gtk_hscale_new(GTK_ADJUSTMENT(yadj));
	gtk_scale_set_draw_value(GTK_SCALE(yhscale), TRUE);
	gtk_scale_set_value_pos(GTK_SCALE(yhscale), GTK_POS_LEFT);
	gtk_scale_set_digits(GTK_SCALE(yhscale), 0);
	gtk_box_pack_start(GTK_BOX(hbox), yhscale, TRUE, TRUE, 5);
	gtk_widget_show(yhscale);
	
	gtk_box_pack_start(GTK_BOX(vbox2),hbox,TRUE,TRUE,4);
	gtk_widget_show(hbox);
	    
	hbox=gtk_hbox_new(FALSE, 4);

	gtk_box_pack_start(GTK_BOX(vbox2),hbox,TRUE,TRUE,4);
	
	gtk_signal_connect(GTK_OBJECT(xadj), "value-changed", 
			   GTK_SIGNAL_FUNC(adj_value_changed), &config.xres);
	gtk_signal_connect(GTK_OBJECT(yadj), "value-changed", 
			   GTK_SIGNAL_FUNC(adj_value_changed), &config.yres);	
	
	gtk_widget_show(hbox);
	

	cbutton = gtk_check_button_new_with_label ("Pixel doubing");
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(cbutton), config.doubling);
	gtk_box_pack_start(GTK_BOX(vbox2), cbutton, FALSE, FALSE, 4);
	gtk_signal_connect(GTK_OBJECT(cbutton), "toggled", GTK_SIGNAL_FUNC(pixeldoubling_toggled),&config.doubling);
	gtk_widget_show(cbutton);

	cbutton = gtk_check_button_new_with_label ("16/9 display");
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(cbutton), config.cinemascope);
	gtk_box_pack_start(GTK_BOX(vbox2), cbutton, FALSE, FALSE, 4);
	gtk_signal_connect(GTK_OBJECT(cbutton), "toggled", GTK_SIGNAL_FUNC(cinemascope_toggled),&config.cinemascope);
	gtk_widget_show(cbutton);

	fsButton = gtk_check_button_new_with_label ("Fullscreen");
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(fsButton), config.fullscreen);
	gtk_box_pack_start(GTK_BOX(vbox2), fsButton, FALSE, FALSE, 4);
	gtk_signal_connect(GTK_OBJECT(fsButton), "toggled", GTK_SIGNAL_FUNC(fullscreen_toggled),&config.fullscreen);
	gtk_widget_show(fsButton);

	
	gtk_widget_show(vbox2);

	gtk_widget_show(frame);

	buttonbox=gtk_hbutton_box_new();
	gtk_box_pack_end(GTK_BOX(vbox), buttonbox, FALSE, FALSE, 8);
	gtk_hbutton_box_set_layout_default(GTK_BUTTONBOX_END);
	gtk_widget_show(buttonbox);
	
	ok=gtk_button_new_with_label("Ok");
	GTK_WIDGET_SET_FLAGS(ok, GTK_CAN_DEFAULT);
	gtk_box_pack_end(GTK_BOX(buttonbox), ok, FALSE, FALSE, 8);
	gtk_widget_show(ok);

	cancel=gtk_button_new_with_label("Cancel");
	GTK_WIDGET_SET_FLAGS(cancel, GTK_CAN_DEFAULT);
	gtk_box_pack_end(GTK_BOX(buttonbox), cancel, FALSE, FALSE, 8);
	gtk_widget_show(cancel);
	
	gtk_widget_show(vbox);

	gtk_window_set_default(GTK_WINDOW(configure_win), ok);
	gtk_signal_connect(GTK_OBJECT(cancel), "clicked", GTK_SIGNAL_FUNC(cancel_clicked), configure_win);
	gtk_signal_connect(GTK_OBJECT(ok), "clicked", GTK_SIGNAL_FUNC(ok_clicked), configure_win);
	
	gtk_widget_show(configure_win);
}

void conf_set_fullscreen (int f) {
  config.fullscreen = f;
  if (configure_win != NULL) {
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(fsButton),
								  config.fullscreen);
  }
  plug_save_prefs ();
}

void plug_save_prefs(void)
{
    char *name;
    ConfigFile *f;
	
	extern_config = config ;
    extern_config.xres = extern_config.xres - (extern_config.xres%32);
    extern_config.yres = extern_config.yres - (extern_config.yres%2);
    
    name=g_strdup_printf("%s%s",g_get_home_dir(), "/.xmms/config");
    
    if(!(f=xmms_cfg_open_file(name)))
	f=xmms_cfg_new();
    
    xmms_cfg_write_string(f, "whatagoom", "version", VERSION) ;
    xmms_cfg_write_int(f, "whatagoom", "xres", config.xres) ;
    xmms_cfg_write_int(f, "whatagoom", "yres", config.yres) ;
	xmms_cfg_write_int(f, "whatagoom", "doubling", config.doubling);
//	xmms_cfg_write_int(f, "whatagoom", "fullscreen", config.fullscreen);
	xmms_cfg_write_int(f, "whatagoom", "cinemascope", config.cinemascope);
    
    xmms_cfg_write_file(f, name);
    xmms_cfg_free(f);

	extern_config = config ;
	extern_config.xres = extern_config.xres - (extern_config.xres%32);
    extern_config.yres = extern_config.yres - (extern_config.yres%2);

/*    if(extern_config.fullscreen)
      {
      extern_config.xres = gdk_screen_width();
      extern_config.yres = gdk_screen_height();
      }
    */
    g_free(name);
}

void plug_load_prefs(void)
{
    char *name;
    ConfigFile *f;
    name=g_strdup_printf("%s%s",g_get_home_dir(), "/.xmms/config");

    if (f = xmms_cfg_open_file(name))
	  {
		if (check_cfg_version(f))
		  {
			xmms_cfg_read_int(f, "whatagoom", "xres", &config.xres);
			xmms_cfg_read_int(f, "whatagoom", "yres", &config.yres);
			xmms_cfg_read_int(f, "whatagoom", "doubling", &config.doubling);
//			xmms_cfg_read_int(f, "whatagoom", "fullscreen", &config.fullscreen);
			xmms_cfg_read_int(f, "whatagoom", "cinemascope", &config.cinemascope);
		  }
		xmms_cfg_free(f);
	  }
    g_free(name);

	extern_config = config ;
    extern_config.xres = extern_config.xres - (extern_config.xres%32);
    extern_config.yres = extern_config.yres - (extern_config.yres%2);

    /* We set the fullscreen resulution here (better than messing up the
     * rest of the code)
     */
    
/*    if(extern_config.fullscreen)
      {
      extern_config.xres = gdk_screen_width();
      extern_config.yres = gdk_screen_height();
      }
*/
}

static gboolean check_cfg_version(ConfigFile *f)
{
    char *vstr;
    
    if(xmms_cfg_read_string(f, "whatagoom", "version", &vstr))
	if(!strcmp(vstr, VERSION))
	    return TRUE;
    
    return FALSE;
}

static void adj_value_changed(GtkWidget *w, int *a)
{
	*a=(int) GTK_ADJUSTMENT(w)->value;
}

static void apply_clicked(GtkWidget *w, GtkWidget *window)
{
	plug_save_prefs();
}

static void cancel_clicked(GtkWidget *w, GtkWidget *window)
{
	gtk_widget_destroy(window);
	configure_win=NULL;
}

static void ok_clicked(GtkWidget *w, GtkWidget *window)
{
    plug_save_prefs();
    gtk_widget_destroy(window);
    configure_win=NULL;
}

/*
 *  fullscreen_toggled and pixeldoubling_toggled could also be the same
 *  I didn't want to mess too much stuff up
 */

static void pixeldoubling_toggled(GtkWidget *w,int *data)
{
  *data=(int) gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(w));
}

static void cinemascope_toggled(GtkWidget *w,int *data)
{
  *data=(int) gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(w));
}

static void fullscreen_toggled(GtkWidget *w,int *data)
{
  *data=(int) gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(w));
}
