#include "config.h"

#include <xmms/plugin.h>  
#include <xmms/xmmsctrl.h>

#include <stdlib.h>
#include <stdio.h>
#include <glib.h>
#include "tinyptc.h"
#include <math.h>

#include "goom_xmms.h"
#include "goom_tools.h"
#include "goom_core.h"

#include "pixeldoubler.h"

static void jeko_init(void);
static void jeko_cleanup(void);
static void jeko_playback_start(void);
static void jeko_playback_stop(void);
static void jeko_render_pcm(gint16 data[2][512]);

void jeko_update_config () ;

VisPlugin jeko_vp =
{
  NULL,
  NULL,
  0,
  "What a GOOM! "VERSION,
  2,
  0,
  jeko_init, /* init */
  jeko_cleanup, /* cleanup */
  about, /* about */
  plug_configure, /* configure */
  NULL, /* disable_plugin */
  jeko_playback_start, /* playback_start */
  jeko_playback_stop, /* playback_stop */
  jeko_render_pcm, /* render_pcm */
  /* jeko_render_freq*/ NULL  /* render_freq */
};

VisPlugin *get_vplugin_info(void) 
{    
  return &jeko_vp;
}   

static Surface * screen = 0;

/*======================*/
/*===============================*/
void jeko_init(void)
{	
  gint16 data[2][512];
  int i;
  int videomode = -1;
  
  guint32 *buftmp = (guint32*)
	malloc (extern_config.xres*extern_config.yres*4);
	
#ifdef VERBOSE
	printf ("--> INITILIZING GOOM\n") ;
#endif
	plug_load_prefs () ;
	jeko_update_config () ;

	/* ouverture */
	if (!ptc_open("What a GOOM!",extern_config.xres,extern_config.yres,
				  extern_config.fullscreen))
	  exit (1);

	ptc_update (buftmp); // to avoid a ptc bug...
	free (buftmp);


	if (extern_config.doubling) {
		screen = surface_new (extern_config.xres,extern_config.yres);
		goom_init (extern_config.xres/2,extern_config.yres/2,extern_config.cinemascope);
//		init_ifs (extern_config.xres/2,extern_config.yres/2);
	}
	else {
		screen = 0;
		goom_init (extern_config.xres,extern_config.yres,extern_config.cinemascope);
//		init_ifs (extern_config.xres,extern_config.yres);
	}
	
	for (i = 0; i < 512; i++) {
	  data[0][i]=0;
	  data[1][i]=0;
	}

	jeko_render_pcm(data);
}

static
void jeko_cleanup(void)
{
#ifdef VERBOSE
	printf ("--> CLEANUP GOOM\n") ;
#endif
	/* fermeture */
	ptc_close();
//	release_ifs ();
	goom_close ();
	if (screen)
	  surface_delete (&screen);
}

static
void jeko_playback_start(void)
{
}

static
void jeko_playback_stop(void)
{
}

static
void jeko_render_pcm(gint16 data[2][512])
{
  static int xres = -1, yres = -1, doubling = -1, fullscreen = -1;
  static Surface * gsurf = NULL;
  static int forceMode = 0;

  if (ptc_key_pressed ('x'))
	xmms_remote_play(jeko_vp.xmms_session);
  if (!fullscreen && (ptc_key_pressed ('c')))
	xmms_remote_pause(jeko_vp.xmms_session);
  if (!fullscreen && (ptc_key_pressed ('v')))
	xmms_remote_stop(jeko_vp.xmms_session);
  if (ptc_key_pressed ('b'))
	xmms_remote_playlist_next(jeko_vp.xmms_session); 
  if (ptc_key_pressed ('z'))
	xmms_remote_playlist_prev(jeko_vp.xmms_session);
  if (ptc_key_pressed ('q'))
	xmms_remote_quit(jeko_vp.xmms_session);
 
  if (fullscreen && ptc_escape_pressed())
	conf_set_fullscreen (0);

  if (!fullscreen && ptc_key_pressed('f'))
	conf_set_fullscreen (1);

  if ((xres != extern_config.xres)
	  || (yres != extern_config.yres)
	  || (doubling != extern_config.doubling)
	  || (fullscreen != extern_config.fullscreen))
	{
//	  int videomode;
//	  int bestx,besty;
//	  if (extern_config.fullscreen) {
//		videomode = ptc_get_nearest_mode (extern_config.xres, extern_config.yres,
//										  &bestx, &besty);
//	  }
		
	  
	  doubling = extern_config.doubling;
	  if (xres != -1)
		{
		  jeko_update_config ();
		  if (ptc_is_open())
			ptc_close ();
		  
		  if (!ptc_open("What a GOOM!",
						extern_config.xres,
						extern_config.yres,
						extern_config.fullscreen))
			exit (1);
		}
	  xres = extern_config.xres;
	  yres = extern_config.yres;
	  fullscreen = extern_config.fullscreen;
	  
	  if (screen) surface_delete (&screen);
	  if (doubling)
		screen = surface_new (extern_config.xres,extern_config.yres);
	  else
		screen = 0;
	  
	  if (gsurf == NULL) {
		gsurf = (Surface*)malloc(sizeof(Surface));
	  }
	  if (doubling) {
		gsurf->width = xres/2;
		gsurf->height = yres/2;
		gsurf->size = xres*yres/4;
	  }
	  else {
		gsurf->width = xres;
		gsurf->height = yres;
		gsurf->size = xres*yres;
	  }
	}
  /* 	if (gsurf == NULL) {
		gsurf = (Surface*)malloc(sizeof(Surface));
		gsurf->width = xres;
		gsurf->height = yres;
		gsurf->size = xres*yres;
		}*/
  forceMode = 0;
  if (ptc_f1_pressed ())
	forceMode = 1;
  if (ptc_f2_pressed ())
	forceMode = 2;
  if (ptc_f3_pressed ())
	forceMode = 3;
  if (ptc_f4_pressed ())
	forceMode = 4;
  if (ptc_f5_pressed ())
	forceMode = 5;
  if (ptc_f6_pressed ())
	forceMode = 6;
  if (ptc_f7_pressed ())
	forceMode = 7;
  if (ptc_f8_pressed ())
	forceMode = 8;
	
  gsurf->buf = goom_update (data,forceMode);
  
  if (doubling) {
	pixel_doubler (gsurf,screen);
	ptc_update (screen->buf);
  }
  else
	ptc_update (gsurf->buf);
}

void jeko_update_config ()
{
  extern_config.xres = extern_config.xres - extern_config.xres % 32;

  if (extern_config.fullscreen) {
	  int x,y,videomode;
	  videomode = ptc_get_nearest_mode (extern_config.xres, extern_config.yres,
										&x,&y);
	  //conf_set_res (x,y);
	  if (videomode != -1) {
		extern_config.xres = x;
		extern_config.yres = y;
	  }
	}

  if (extern_config.doubling) {
  	goom_set_resolution (extern_config.xres/2, extern_config.yres/2,extern_config.cinemascope) ;
//	init_ifs (extern_config.xres/2, extern_config.yres/2);
  }
  else {
	goom_set_resolution (extern_config.xres, extern_config.yres,extern_config.cinemascope);
//	init_ifs (extern_config.xres, extern_config.yres);
  }
}
