#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <Evas.h>
#include <Ecore.h>

#include "goom_tools.h"
#include "goom_core.h"

#include "ifs.h"

float width = 320;
float height = 240;
int use_bg = 1;

/*
 * internal config
 */
#define RENDER_ENGINE RENDER_METHOD_ALPHA_SOFTWARE

#define MAX_EVAS_COLORS (216)
#define MAX_FONT_CACHE (512 * 1024)
#define MAX_IMAGE_CACHE (1 * (1024 * 1024))

#define FONT_DIRECTORY "./fnt"

Evas_Render_Method render_method = RENDER_ENGINE;

/*
 * Output object.
 */
Evas evas;
Window main_window;

/*
 * Goom interface
 */
gint16 snd_data [2][512]; /* sound for goom */
int * data;               /* display of goom */

/*
 * Les objets Evas
 */
Evas_Object o_image; /* the goom surface */
Evas_Object o_text;  /* fps display */
Evas_Object o_bg;    /* the background image */

/*
 * For fps calculation
 */
struct timeval timev_init; /* time at program start */
long t_count = 0;          /* nb frame displayed */


/*
 * Force a Goom Mode
 */
static int forceMode = 0;
static int doForceMode = 0;

/** return the nb of sec for a timeval */
double get_seconds (struct timeval tv) {
  return (double)tv.tv_sec + (((double)tv.tv_usec) / 1000000.0); }


/*
 * Timer
 */
static void
e_timeout (int val, void *not_used)
{
  t_count ++;
  ecore_add_event_timer ("e_timeout()", 0.01, e_timeout, val+1, NULL);
}


static void e_key_down (Ecore_Event *ev) {
	forceMode = (forceMode % NB_FX) + 1;
	doForceMode = 1;
	printf ("Switch to mode %d\n", forceMode);
}

/*
 * Main Method
 */
static void
e_idle (void *not_used)
{
  int i;
  char texte [255];
  static int count = 0;
  static int monbool = 1;

  IFSPoint * points;
  int nbpt;

  while (t_count > count) {
    count ++;

    /** CHANGEMENT DU SON EMULE **/
    if (rand () % 25 == 0) {
      if ((monbool = rand () % 3) == 0) {
	for (i = 0; i < 512; i++) {
	  snd_data [0][i] = 0;
	  snd_data [1][i] = 0;
	}
      }
    }
    if (monbool) {
      for (i = 0; i < 512; i++) {
	snd_data [0][i] = rand();
	snd_data [1][i] = rand();
      }
    }
    
    /** MISE A JOUR DES BUFFERS DE GOOM **/
    data = goom_update (snd_data, forceMode*doForceMode);
		doForceMode = 0;
//    ifs_update (data, width, height, 1);
    
    /** MISE A JOUR DU BUFFER DE L'IMAGE EVAS **/
    evas_set_image_data (evas, o_image, data, IMAGE_FORMAT_BGRA, width, height);
    
    /* calcul du fps */
    {
      struct timeval tv;
      gettimeofday (&tv, NULL);
      timersub (&tv, &timev_init, &tv);
      sprintf (texte, "%2.0f fps", (double)(count) / get_seconds (tv));
      evas_set_text (evas, o_text, texte);
    }
  }
  evas_render (evas);
}


/*
 * exposure 
 */
static void
e_window_expose (Ecore_Event * ev) {
  Ecore_Event_Window_Expose *e;
  e = (Ecore_Event_Window_Expose *) ev->event;
  evas_update_rect (evas, e->x, e->y, e->w, e->h);
}


/*
 * resizing
 */
static void
e_window_configure (Ecore_Event *ev)
{
  Ecore_Event_Window_Configure *e;
  e = (Ecore_Event_Window_Configure *) ev->event;

  if (e->win == main_window) {
    ecore_window_resize (evas_get_window(evas), e->w, e->h);
    evas_set_output_size (evas, e->w, e->h);
  }
}


/*
 * Evas setup.
 */
void
setup () {
  int i;
  Window ewin;

  /* setup callback for events */
  ecore_event_filter_handler_add (ECORE_EVENT_WINDOW_EXPOSE,
																	e_window_expose);
	
  ecore_event_filter_handler_add (ECORE_EVENT_WINDOW_CONFIGURE,
																	e_window_configure);

	ecore_event_filter_handler_add (ECORE_EVENT_KEY_DOWN, e_key_down);

  /* handler for when the event queue goes idle */
  ecore_event_filter_idle_handler_add (e_idle, NULL);

  main_window = ecore_window_new (0,0,0,width,height);
  evas = evas_new_all (ecore_display_get(), main_window,
											 0,0, width, height, render_method,
											 MAX_EVAS_COLORS, MAX_FONT_CACHE, MAX_IMAGE_CACHE,
											 FONT_DIRECTORY);
  
  /* get the window ID for the evas created for us */
  ewin = evas_get_window (evas);
	
  /* show it */
  ecore_window_show (ewin);
  /* set the event this window accepts */
  ecore_window_set_events (ewin, XEV_EXPOSE|XEV_KEY);
  /* show the toplevel */
  ecore_window_set_title(main_window, "What a Goom!");

  //  ecore_window_hint_set_borderless(main_window);
  //ecore_window_move_resize (main_window,0,0,640,480);
  //  ecore_window_raise (main_window);

  ecore_window_show (main_window);

  /** EXTERNAL MODULE INIT **/

  goom_init (width, height, 0);
  data = goom_update (snd_data,0);
//  init_ifs (width, height);

  /****** NOW CREATE OBJECTS IN THE EVAS *******/
  //  for (i = 0; i < width*height; i++) data [i] = rand();

if (use_bg) {
  o_bg = evas_add_image_from_file (evas, "image.jpg");
  evas_resize (evas, o_bg, width,height);
  evas_set_image_fill (evas, o_bg, 0.0,0.0,width,height);
  evas_move (evas, o_bg, 0.0, 0.0);
  evas_show (evas, o_bg);
}

  o_image = evas_add_image_from_data (evas, data, IMAGE_FORMAT_BGRA, width, height);
 
  if (use_bg)
    evas_set_color (evas, o_image, 255,255,255,192);
  evas_move (evas, o_image, 0.0, 0.0);
  evas_show (evas, o_image);
  
  o_text = evas_add_text (evas, "notepad", 10, "Evas Rulez Da World !!!");
  evas_move (evas, o_text, 5.0, 5.0);
  evas_set_color (evas, o_text, 255,255,255,255);
  evas_show (evas, o_text);
}


/*
 * Initialisation of all that need to be.
 */
int
main (int argc, char **argv) {
  printf ("baffe version 1.0 by JeKo\n");
  printf ("usage : baffe [options]\n"
	  "  -res x y       change the size of the window\n"
	  "  -soft          use the software rendering engine\n"
	  "  -hard          use a hardware rendering engine (OpenGL)\n"
	  "  -x11           use the X11 rendering engine\n"
	  "  -nobg          dont display a background behind the effect\n"
	  "\n");

  /* command line parsing */
  {
    int i;
    for (i = 0; i < argc; i++) {
      if (!strcmp (argv[i], "-soft"))
	render_method = RENDER_METHOD_ALPHA_SOFTWARE;
      if (!strcmp (argv[i], "-x11"))
	render_method = RENDER_METHOD_BASIC_HARDWARE;
      if (!strcmp (argv[i], "-hard"))
	render_method = RENDER_METHOD_3D_HARDWARE;
      if (!strcmp (argv[i], "-nobg"))
	use_bg = 0;;
      if (!strcmp (argv[i], "-res")) {
	if (argc > i+2) {
	  sscanf (argv[i+1], "%f", &width);
	  sscanf (argv[i+2], "%f", &height);
	}
	else {
	  printf ("Specify the resolution please\n");
	}
      }
    }
  }

  /* init X */
  ecore_display_init (NULL);
  /* setup handlers for system signals */
  ecore_event_signal_init ();
  /* setup the X event internals */
  ecore_event_x_init ();

  /* data setup */
  setup ();

  gettimeofday (&timev_init, NULL);

  e_timeout (0, NULL);
  /* event loop (for ever) */
  ecore_event_loop ();

  return 0;
}
