/*
 * 
 * TinyPTC x11 v0.5.1 ported by Alessandro Gatti (a.gatti@tiscalinet.it)
 * Original Windows version by Glenn Fiedler (ptc@gaffer.org)
 * http://www.gaffer.org/tinyptc
 * 
 */

#include "config.h"
#include "tinyptc.h"

#ifdef __PTC_ENABLE_CONVERSIONS__
#include "convert.h"
#endif /* __PTC_ENABLE_CONVERSIONS__ */

/* #defines' sanity checks */

#ifdef __PTC_XSHM__
#ifndef __PTC_WINDOWED__
#define __PTC_WINDOWED__
#endif /* !__PTC_WINDOWED__ */
#endif /* __PTC_XSHM__ */

#ifndef __PTC_WINDOWED__
#define __PTC_DGA__
#endif /* !__PTC_WINDOWED__ */

/* Common #defines */

#define __PTC_FUNCTION_KEY__ 0xFF
#define __PTC_ESCAPE_KEY__ 0x1B
#define __PTC_F1_KEY__ 0xBE
#define __PTC_F2_KEY__ 0xBF
#define __PTC_F3_KEY__ 0xC0
#define __PTC_F4_KEY__ 0xC1
#define __PTC_F5_KEY__ 0xC2
#define __PTC_F6_KEY__ 0xC3
#define __PTC_F7_KEY__ 0xC4
#define __PTC_F8_KEY__ 0xC5
#define __PTC_A_KEY__ 0x61
#define __PTC_F_KEY__ 0x66

/* #includes */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>

/** test by JeKo **/
#include <X11/extensions/xf86vmode.h>
#include <math.h>


#ifdef __PTC_DGA__
#include <X11/extensions/xf86dga.h>
#include <X11/extensions/xf86vmode.h>
#include <unistd.h>
#include <sys/types.h>
#endif /* __PTC_DGA__ */

#ifdef __PTC_XSHM__
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#endif /* __PTC_XSHM__ */

#define PTC_FAILURE 0
#define PTC_SUCCESS 1

#ifdef __PTC_CLEANUP_CALLBACK__
extern void ptc_cleanup_callback (void);
#endif /* __PTC_CLEANUP_CALLBACK__ */

static Display *ptc_display;
static int ptc_screen, ptc_screen_width, ptc_screen_height,
  ptc_viewport_width, ptc_viewport_height, ptc_depth;
static Window ptc_root_window;
static Visual *ptc_visual;
static char *ptc_buffer;
static int ptc_pixmap_format_count, ptc_pixmap_counter, ptc_converter_depth;
XPixmapFormatValues *ptc_pixmap_formats;

#ifdef __PTC_WINDOWED__
static Window ptc_window;
static int ptc_x_position, ptc_y_position;
static XSetWindowAttributes ptc_window_attributes;
static XSizeHints ptc_window_sizehints;
static GC ptc_window_gc;
static XImage *ptc_ximage;
#endif /* __PTC_WINDOWED__ */

#ifdef __PTC_XSHM__
static XShmSegmentInfo ptc_shm_segment;
#endif /* __PTC_XSHM__ */

#ifdef __PTC_DGA__
static char *ptc_framebuffer_address;
static int ptc_framebuffer_width, ptc_framebuffer_banksize,
  ptc_framebuffer_memory, ptc_framebuffer_start, ptc_framebuffer_pitch,
  ptc_framebuffer_index, ptc_totalmodes, ptc_dotclock, ptc_modecounter,
  ptc_previousmode_number, ptc_event_base, ptc_error_base, ptc_blitcounter;
static Bool ptc_emulate_fullscreen;
static XF86VidModeModeInfo **ptc_modeinfodata;
static XF86VidModeModeLine ptc_currentmodeline;
#ifdef __PTC_BEST_VIDEOMODE__
static int ptc_best_x_delta, ptc_best_y_delta, ptc_best_videomode,
  ptc_best_delta;
#endif /* __PTC_BEST_VIDEOMODE__ */
#endif /* __PTC_DGA__ */

#ifdef __PTC_ENABLE_CONVERSIONS__
PTC_CONVERTER ptc_convert;
static int ptc_output_pitch, ptc_blitcounter, ptc_source_index,
  ptc_destination_index;
#endif

static int escape_pressed = 0;
static int f1_pressed = 0;
static int f2_pressed = 0;
static int f3_pressed = 0;
static int f4_pressed = 0;
static int f5_pressed = 0;
static int f6_pressed = 0;
static int f7_pressed = 0;
static int f8_pressed = 0;
static int key_pressed [256];

static int ptc_opened = 0;
static int ptc_previousmode_number = -1; /* ancien mode xvidmode */
static int ptc_previousscreen_width = -1;
/*
 *   Make a window fullscreen (Mwm stuff taken from gdk)
 */

static Cursor fullscreen_cursor        = None;
static Pixmap fullscreen_cursor_pixmap = None;
static Pixmap fullscreen_cursor_mask   = None;

typedef struct {
    unsigned long flags;
    unsigned long functions;
    unsigned long decorations;
    long input_mode;
    unsigned long status;
} MotifWmHints, MwmHints;

#define MWM_HINTS_FUNCTIONS     (1L << 0)
#define MWM_HINTS_DECORATIONS   (1L << 1)
#define MWM_HINTS_INPUT_MODE    (1L << 2)
#define MWM_HINTS_STATUS        (1L << 3)

#define MWM_FUNC_ALL            (1L << 0)
#define MWM_FUNC_RESIZE         (1L << 1)
#define MWM_FUNC_MOVE           (1L << 2)
#define MWM_FUNC_MINIMIZE       (1L << 3)
#define MWM_FUNC_MAXIMIZE       (1L << 4)
#define MWM_FUNC_CLOSE          (1L << 5)

#define MWM_DECOR_ALL           (1L << 0)
#define MWM_DECOR_BORDER        (1L << 1)
#define MWM_DECOR_RESIZEH       (1L << 2)
#define MWM_DECOR_TITLE         (1L << 3)
#define MWM_DECOR_MENU          (1L << 4)
#define MWM_DECOR_MINIMIZE      (1L << 5)
#define MWM_DECOR_MAXIMIZE      (1L << 6)
#define _XA_MOTIF_WM_HINTS              "_MOTIF_WM_HINTS"

static MwmHints fullscreen_hints =
  {
    flags: MWM_HINTS_DECORATIONS,
    functions: 0,   /* Unsued */
    decorations: 0,
    input_mode: 0,  /* Unused */
    status: 0,      /* Unused */
  };

static unsigned char cursor_data[] = { 0x00 };
static unsigned char cursor_mask[] = { 0x00 };

static XF86VidModeModeInfo **ptc_modeinfodata = NULL;
static int ptc_totalmodes = 0;

static int ptc_set_xvidmode (int mode)
{
  static XF86VidModeModeLine ptc_currentmodeline;
  static int ptc_event_base, ptc_error_base, ptc_totalmodes, ptc_dotclock;
  int ptc_modecounter;

#ifdef VERBOSE
  printf ("trying to set mode %d\n", mode);
#endif
  if (mode == -1)
	return -1;

#ifdef VERBOSE
  printf ("Check for VidMode extension avaliability\n");
#endif
  if (!XF86VidModeQueryExtension
	  (ptc_display, &ptc_event_base, &ptc_error_base))
	  {
#ifdef VERBOSE
		printf ("couldn't change video mode..\n");
#endif
		return -1;
	  }
  
#ifdef VERBOSE
  printf ("Get all the mode parameters\n");
#endif
//  if (ptc_modeinfodata == NULL){
	XF86VidModeGetAllModeLines (ptc_display, ptc_screen, &ptc_totalmodes,
								&ptc_modeinfodata);
//  }
  
#ifdef VERBOSE
  printf ("nb mode = %d\n", ptc_totalmodes);
#endif
  if (mode >= ptc_totalmodes)
	return -1;
  
#ifdef VERBOSE
  printf("Get current mode parameters\n");
#endif
  XF86VidModeGetModeLine (ptc_display, ptc_screen, &ptc_dotclock,
						  &ptc_currentmodeline);
  
#ifdef VERBOSE
  printf ("Search for current mode in parameter list\n");
#endif
  for (ptc_modecounter = 0; ptc_modecounter < ptc_totalmodes;
	   ptc_modecounter++) {
	if ((ptc_currentmodeline.hdisplay ==
		 ptc_modeinfodata[ptc_modecounter]->hdisplay)
		&& (ptc_currentmodeline.vdisplay ==
			ptc_modeinfodata[ptc_modecounter]->vdisplay))
	  {
#ifdef VERBOSE
		printf("Save the parameters\n");
#endif
		ptc_previousmode_number = ptc_modecounter;
		ptc_previousscreen_width = ptc_currentmodeline.hdisplay;
		break;
	  }
  }
  
#ifdef VERBOSE
  printf ("switching to %d\n", mode);
#endif
  XF86VidModeSwitchToMode(ptc_display,
						  ptc_screen,
						  ptc_modeinfodata [mode]);
  
  	XF86VidModeSetViewPort (ptc_display,
  							ptc_screen,
  							0,0);
  
  //  XFree (ptc_modeinfodata);
  return 0;
}

static int ptc_restore_xvidmode ()
{
  static XF86VidModeModeInfo **ptc_modeinfodata;
  static XF86VidModeModeLine ptc_currentmodeline;
  static int ptc_event_base, ptc_error_base, ptc_totalmodes, ptc_dotclock;
  int ptc_modecounter;

#ifdef VERBOSE
  printf("Check for VidMode extension avaliability");
#endif
  if (!XF86VidModeQueryExtension
	  (ptc_display, &ptc_event_base, &ptc_error_base))
	{
#ifdef VERBOSE
	  fprintf (stderr, "couldn't change video mode..\n");
#endif
	  return -1;
	}
  else {
	/* Get all the mode parameters */
//	if (ptc_modeinfodata == NULL)
	  XF86VidModeGetAllModeLines (ptc_display, ptc_screen, &ptc_totalmodes,
								  &ptc_modeinfodata);
	
	
	/* Search for current mode in parameter list */
	for (ptc_modecounter = 0; ptc_modecounter < ptc_totalmodes;
		 ptc_modecounter++) {
	  if ((ptc_previousscreen_width ==
		   ptc_modeinfodata[ptc_modecounter]->hdisplay)
		  && (ptc_currentmodeline.vdisplay ==
			  ptc_modeinfodata[ptc_modecounter]->vdisplay))
		{
		  /* Save the parameters */
		  XF86VidModeSwitchToMode(ptc_display,
								  ptc_screen,
								  ptc_modeinfodata [ptc_modecounter]);
		  break;
		}
	}
  }
//  XFree (ptc_modeinfodata);
}

/* return -1 if not success : a VideoModeNumber else */
int ptc_get_nearest_mode (int width, int height, int*nwidth, int *nheight) {

//  static XF86VidModeModeInfo **ptc_modeinfodata;
  static XF86VidModeModeLine ptc_currentmodeline;
  static int ptc_event_base, ptc_error_base, ptc_totalmodes, ptc_dotclock;
  int ptc_modecounter;
  int mindistx = 0x7fffffff;
  int bestmode = -1;
  Display * tmp_display;
  int tmp_screen;

#ifdef VERBOSE
  printf ("scearching nearest frm %dx%d\n",width,height);
#endif

  tmp_display = XOpenDisplay (NULL);
  if (tmp_display == NULL)
    {
      return -1;
    }
  /* Get the default screen associated with the previously opened display */
  tmp_screen = DefaultScreen (tmp_display);

  /* Check for VidMode extension avaliability */
  if (!XF86VidModeQueryExtension
	  (tmp_display, &ptc_event_base, &ptc_error_base))
	{
#ifdef VERBOSE
	  fprintf (stderr, "couldn't change video mode..\n");
#endif
	  return -1;
	}
  else {
	/* Get all the mode parameters */
//	if (ptc_modeinfodata == NULL)
	  XF86VidModeGetAllModeLines (tmp_display, tmp_screen, &ptc_totalmodes,
								  &ptc_modeinfodata);
	
	/* Get current mode parameters */
	//	  XF86VidModeGetModeLine (ptc_display, ptc_screen, &ptc_dotclock,
	//							  &ptc_currentmodeline);
	
	/* Search for current mode in parameter list */
	for (ptc_modecounter = 0; ptc_modecounter < ptc_totalmodes;
		 ptc_modecounter++) {
	  int delta = width - ptc_modeinfodata[ptc_modecounter]->hdisplay;
	  if (delta < 0) delta = -delta;
#ifdef VERBOSE
	  printf ("mode %d : %d - %d => %d\n", ptc_modecounter,
			  ptc_modeinfodata[ptc_modecounter]->hdisplay, width,
			  delta);
#endif
	  if (delta < mindistx) {
		bestmode = ptc_modecounter;
		mindistx = delta;
	  }
	}

	if (bestmode == -1)
	  return -1;

	*nwidth = ptc_modeinfodata[bestmode]->hdisplay;
	/** BUG ??? BUG ??? BUG ??? **/
		if (ptc_modeinfodata[bestmode]->vsyncstart * 4 / 3
			== ptc_modeinfodata[bestmode]->hdisplay)
	*nheight = ptc_modeinfodata[bestmode]->vsyncstart;
		else
		  *nheight = ptc_modeinfodata[bestmode]->vdisplay;
	XCloseDisplay (tmp_display);
//	XFree (ptc_modeinfodata);
  
	return bestmode;
  }
}

static void make_fullscreen_window()
{
  static Atom hints_atom = None;
  MotifWmHints *hints;
  Atom type;
  int format;
  unsigned long nitems;
  unsigned long bytes_after;

  /* Stuff for making the cursor */

  Colormap cmap;
  unsigned long black_pixel, white_pixel;
  XColor black_color;
  XColor white_color;
  Window root_window;  
  
  /*
   *   Set the decorations (via Mwm hints)
   *   This is taken from the sources of gdk-1.2.8
   */
  
  if (!hints_atom)
    hints_atom = XInternAtom (ptc_display, 
                              _XA_MOTIF_WM_HINTS, False);
  
  XGetWindowProperty (ptc_display, ptc_window,
                      hints_atom, 0, sizeof (MotifWmHints)/sizeof (long),
                      False, AnyPropertyType, &type, &format, &nitems,
                      &bytes_after, (unsigned char **)&hints);
  
  if (type == None)
    hints = &fullscreen_hints;
  else
    {
      if (fullscreen_hints.flags & MWM_HINTS_FUNCTIONS)
        {
          hints->flags |= MWM_HINTS_FUNCTIONS;
          hints->functions = fullscreen_hints.functions;
        }
      if (fullscreen_hints.flags & MWM_HINTS_DECORATIONS)
        {
          hints->flags |= MWM_HINTS_DECORATIONS;
          hints->decorations = fullscreen_hints.decorations;
        }
    }
  
  XChangeProperty (ptc_display, ptc_window,
                   hints_atom, hints_atom, 32, PropModeReplace,
                   (unsigned char *)hints,
                   sizeof (MotifWmHints)/sizeof (long));
  
  if (hints != &fullscreen_hints)
    XFree (hints);

  /*
   *  Make the fullscreen cursor (ported from my previous hacked
   *  goom-1.7.6, which was, in turn, ported from the fullscreen video
   *  mechanisms of gmerlin)
   */

  black_pixel = BlackPixel(ptc_display, ptc_screen);
  white_pixel = WhitePixel(ptc_display, ptc_screen);
  
  root_window = RootWindow(ptc_display, ptc_screen);
  cmap = DefaultColormap(ptc_display, ptc_screen);

  black_color.pixel = black_pixel;
  white_color.pixel = white_pixel;
  XQueryColor(ptc_display, cmap, &black_color);
  XQueryColor(ptc_display, cmap, &white_color);
  
  fullscreen_cursor_pixmap =
    XCreatePixmapFromBitmapData(ptc_display, root_window, 
                                cursor_data, 1, 1, black_pixel,
                                white_pixel, 1);
  
  fullscreen_cursor_mask =
    XCreatePixmapFromBitmapData(ptc_display, root_window,
                                cursor_mask, 1, 1, black_pixel,
                                white_pixel, 1);
  
  fullscreen_cursor = XCreatePixmapCursor(ptc_display,
                                          fullscreen_cursor_pixmap,
                                          fullscreen_cursor_mask,
                                          &black_color, &white_color, 0, 0);
  XDefineCursor(ptc_display, ptc_window, fullscreen_cursor);
  
  XWarpPointer(ptc_display, None, ptc_window, 0, 0, ptc_screen_width, ptc_screen_height, 0,0);
}

int ptc_f1_pressed () {
  int ret = f1_pressed;  f1_pressed = 0;  return ret; }
int ptc_f2_pressed () {
  int ret = f2_pressed;  f2_pressed = 0;  return ret; }
int ptc_f3_pressed () {
  int ret = f3_pressed;  f3_pressed = 0;  return ret; }
int ptc_f4_pressed () {
  int ret = f4_pressed;  f4_pressed = 0;  return ret; }
int ptc_f5_pressed () {
  int ret = f5_pressed;  f5_pressed = 0;  return ret; }
int ptc_f6_pressed () {
  int ret = f6_pressed;  f6_pressed = 0;  return ret; }
int ptc_f7_pressed () {
  int ret = f7_pressed;  f7_pressed = 0;  return ret; }
int ptc_f8_pressed () {
  int ret = f8_pressed;  f8_pressed = 0;  return ret; }
int ptc_key_pressed (char c) {
  int ret = key_pressed[c];  key_pressed[c] = 0;  return ret; }
int ptc_escape_pressed () {
  int ret = escape_pressed;
  escape_pressed = 0;
  return ret;
}

int ptc_is_open () {
  return ptc_opened;
}

/* Open the screen */

int
ptc_open (char *title, int width, int height,
		  int fullscreen)
{
  int i;
  escape_pressed = 0;
  f1_pressed = 0;
  f2_pressed = 0;
  f3_pressed = 0;
  f4_pressed = 0;
  f5_pressed = 0;
  f6_pressed = 0;
  f7_pressed = 0;
  f8_pressed = 0;
  for (i=0;i<256;i++)
	key_pressed [i] = 0;
#ifdef __PTC_DGA__
  /* Check for the effective UID of the running process */
  if (geteuid ())
    {
      return PTC_FAILURE;
    }
#endif /* __PTC_DGA__ */
  /* Open a display on the current root window */
  ptc_display = XOpenDisplay (NULL);
  if (ptc_display == NULL)
    {
      return PTC_FAILURE;
    }
  /* Get the default screen associated with the previously opened display */
  ptc_screen = DefaultScreen (ptc_display);
  /* Get the default visual */
  ptc_visual = DefaultVisual (ptc_display, ptc_screen);
  /* Get screen bitdepth */
  ptc_depth = DefaultDepth (ptc_display, ptc_screen);
  /* Get a pointer to the supported pixmap formats */
  ptc_pixmap_formats =
    XListPixmapFormats (ptc_display, &ptc_pixmap_format_count);
  /* Check if there's one that's suitable */
  for (ptc_pixmap_counter = 0; ptc_pixmap_counter < ptc_pixmap_format_count;
       ptc_pixmap_counter++)
    {
      if (ptc_depth == ptc_pixmap_formats[ptc_pixmap_counter].depth)
	{
	  /* Set the right value */
	  ptc_converter_depth =
	    ptc_pixmap_formats[ptc_pixmap_counter].bits_per_pixel;
	}
    }
  XFree (ptc_pixmap_formats);
#ifdef __PTC_ENABLE_CONVERSIONS__
  /* Check if a converter is avaliable */
  ptc_convert =
    ptc_request_converter (ptc_converter_depth, ptc_visual->red_mask,
			   ptc_visual->green_mask, ptc_visual->blue_mask);
  if (!ptc_convert)
    {
      /* Close the display */
      XCloseDisplay (ptc_display);
      return 0;
    }
  /* Get the actual bytes-per-pixel value */
  switch (ptc_converter_depth)
    {
    case 8:
      ptc_output_pitch = 1;
      break;
    case 15:
      ptc_output_pitch = 2;
      break;
    case 16:
      ptc_output_pitch = 2;
      break;
    case 24:
      ptc_output_pitch = 3;
      break;
    case 32:
      ptc_output_pitch = 4;
      break;
    }
  /* Allocate the temporary buffer */
  ptc_buffer = (char *) malloc (width * height * ptc_output_pitch);
  if (ptc_buffer == NULL)
    {
      XCloseDisplay (ptc_display);
      return PTC_FAILURE;
    }
#else
  /* It runs only on a 32bpp display if no conversions were activated */
  if (ptc_converter_depth != 32)
    {
      XCloseDisplay (ptc_display);
      return PTC_FAILURE;
    }
#endif /* __PTC_ENABLE_CONVERSIONS__ */
#ifdef __PTC_DGA__
  /* Check for DGA extension avaliability */
  if (!XF86DGAQueryExtension (ptc_display, &ptc_event_base, &ptc_error_base))
    {
      XCloseDisplay (ptc_display);
      return PTC_FAILURE;
    }
  /* Get framebuffer parameters */
  XF86DGAGetVideo (ptc_display, ptc_screen, &ptc_framebuffer_address,
		   &ptc_framebuffer_width, &ptc_framebuffer_banksize,
		   &ptc_framebuffer_memory);
  /* Check if the mode is banked */
  if (ptc_framebuffer_banksize < ptc_framebuffer_memory)
    {
      XCloseDisplay (ptc_display);
      return PTC_FAILURE;
    }
  /* Check for VidMode extension avaliability */
  if (!XF86VidModeQueryExtension
      (ptc_display, &ptc_event_base, &ptc_error_base))
    {
      XCloseDisplay (ptc_display);
      return PTC_FAILURE;
    }
  /* Get all the mode parameters */
  XF86VidModeGetAllModeLines (ptc_display, ptc_screen, &ptc_totalmodes,
			      &ptc_modeinfodata);
  /* Get current mode parameters */
  XF86VidModeGetModeLine (ptc_display, ptc_screen, &ptc_dotclock,
			  &ptc_currentmodeline);
  /* Search for current mode in parameter list */
  for (ptc_modecounter = 0; ptc_modecounter < ptc_totalmodes;
       ptc_modecounter++)
    {
      if (
	  (ptc_currentmodeline.hdisplay ==
	   ptc_modeinfodata[ptc_modecounter]->hdisplay)
	  && (ptc_currentmodeline.vdisplay ==
	      ptc_modeinfodata[ptc_modecounter]->vdisplay))
	{
	  /* Save the parameters */
	  ptc_previousmode_number = ptc_modecounter;
	  ptc_screen_width = ptc_currentmodeline.hdisplay;
	  ptc_screen_height = ptc_currentmodeline.vdisplay;
	  break;
	}
    }
  ptc_emulate_fullscreen = True;
  /* Search for desired mode in parameter list */
  for (ptc_modecounter = 0; ptc_modecounter < ptc_totalmodes;
       ptc_modecounter++)
    {
      if ((ptc_modeinfodata[ptc_modecounter]->hdisplay == width)
	  && (ptc_modeinfodata[ptc_modecounter]->vdisplay == height))
	{
	  /* Try to switch mode */
	  if (!XF86VidModeSwitchToMode
	      (ptc_display, ptc_screen, ptc_modeinfodata[ptc_modecounter]))
	    {
	      XFree (ptc_modeinfodata);
	      XCloseDisplay (ptc_display);
	      return PTC_FAILURE;
	    }
	  ptc_emulate_fullscreen = False;
	  break;
	}
    }
#ifdef __PTC_BEST_VIDEOMODE__
  /* It works for fullscreen emulation only */
  if (ptc_emulate_fullscreen)
    {
      /* Set up loop */
      ptc_best_delta = INT_MAX;
      ptc_best_videomode = 0;
      /* Check for all avaliable modes */
      for (ptc_modecounter = 0; ptc_modecounter < ptc_totalmodes;
	   ptc_modecounter++)
	{
	  if ((ptc_modeinfodata[ptc_modecounter]->hdisplay >= width)
	      && (ptc_modeinfodata[ptc_modecounter]->vdisplay >= height))
	    {
	      /* If it fits, get delta parameters */
	      ptc_best_x_delta =
		ptc_modeinfodata[ptc_modecounter]->hdisplay - width;
	      ptc_best_x_delta *= ptc_best_x_delta;
	      ptc_best_y_delta =
		ptc_modeinfodata[ptc_modecounter]->vdisplay - height;
	      ptc_best_y_delta *= ptc_best_y_delta;
	      /* Check if the mode was fitting better than the previous one */
	      if (ptc_best_x_delta + ptc_best_y_delta < ptc_best_delta)
		{
		  /* Set up next iteration */
		  ptc_best_delta = ptc_best_x_delta + ptc_best_y_delta;
		  ptc_best_videomode = ptc_modecounter;
		}
	    }
	}
      /* Try to switch mode */
      if (!XF86VidModeSwitchToMode
	  (ptc_display, ptc_screen, ptc_modeinfodata[ptc_best_videomode]))
	{
	  XFree (ptc_modeinfodata);
	  XCloseDisplay (ptc_display);
	  return PTC_FAILURE;
	}
    }
#endif /* __PTC_BEST_VIDEOMODE__ */
  /* Clear event queue */
  XFlush (ptc_display);
  /* Get the root window */
  ptc_root_window = XRootWindow (ptc_display, ptc_screen);
  /* Get exclusive keyboard access */
  XGrabKeyboard (ptc_display, ptc_root_window, False, GrabModeAsync,
		 GrabModeAsync, CurrentTime);
  /* Get exclusive mouse access */
  XGrabPointer (ptc_display, ptc_root_window, True,
		PointerMotionMask | ButtonPressMask | ButtonReleaseMask,
		GrabModeAsync, GrabModeAsync, None, None, CurrentTime);
  /* Leave root mode */
  setuid (getuid ());
  /* Get viewport parameters */
  XF86DGAGetViewPortSize (ptc_display, ptc_screen,
			  (int *) (&ptc_screen_width),
			  (int *) (&ptc_screen_height));
  /* Set fullscreen emulation up */
  if (ptc_emulate_fullscreen)
    {
      /* Get effective framebuffer address */
#ifdef __PTC_ENABLE_CONVERSIONS__
      ptc_framebuffer_start =
	ptc_framebuffer_width * ((ptc_screen_height - height) / 2) *
	ptc_output_pitch +
	((ptc_screen_width - width) / 2) * ptc_output_pitch;
#else
      ptc_framebuffer_start =
	ptc_framebuffer_width * ((ptc_screen_height - height) / 2) *
	sizeof (int) + ((ptc_screen_width - width) / 2) * sizeof (int);
#endif /* __PTC_ENABLE_CONVERSIONS__ */
    }
  else
    {
      /* Ignore offsets */
      ptc_framebuffer_start = 0;
    }
  /* Get the pitch value */
#ifdef __PTC_ENABLE_CONVERSIONS__
  ptc_framebuffer_pitch = ptc_framebuffer_width * ptc_output_pitch;
#else
  ptc_framebuffer_pitch = ptc_framebuffer_width * sizeof (int);
#endif /* __PTC_ENABLE_CONVERSIONS__ */
  /* Do a safety fork */
  if (XF86DGAForkApp (ptc_screen))
    {
      /* Exit DGA mode */
      XF86DGADirectVideo (ptc_display, ptc_screen, False);
      /* Switch back to the previous video mode */
      XF86VidModeSwitchToMode (ptc_display, ptc_screen,
			       ptc_modeinfodata[ptc_previousmode_number]);
      /* Free modeinfo data */
      XFree (ptc_modeinfodata);
      /* Leave exclusive keyboard access */
      XUngrabKeyboard (ptc_display, CurrentTime);
      /* Leave exclusive mouse access */
      XUngrabPointer (ptc_display, CurrentTime);
      /* Close the display */
      XCloseDisplay (ptc_display);
      return PTC_FAILURE;
    }
  /* Initialize DGA video access */
  XF86DGADirectVideo (ptc_display, ptc_screen,
		      XF86DGADirectGraphics | XF86DGADirectKeyb |
		      XF86DGADirectMouse);
  /* Clear the screen */
#ifdef __PTC_ENABLE_CONVERSIONS__
  memset (ptc_framebuffer_address, 0x00,
	  ptc_framebuffer_width * ptc_screen_height * ptc_output_pitch);
#else
  memset (ptc_framebuffer_address, 0x00,
	  ptc_framebuffer_width * ptc_screen_height * sizeof (int));
#endif /* __PTC_ENABLE_CONVERSIONS__ */
  /* Select the input events that should be reported */
  XSelectInput (ptc_display, DefaultRootWindow (ptc_display),
		KeyPressMask | KeyReleaseMask);
  /* Set the final viewport */
  XF86DGASetViewPort (ptc_display, ptc_screen, 0, 0);
  /* Clear event queue */
  XFlush (ptc_display);
  /* Save the buffer size */
  ptc_viewport_width = width;
  ptc_viewport_height = height;
#endif /* __PTC_DGA__ */
#ifdef __PTC_WINDOWED__
#ifdef __PTC_XSHM__
  /* Check for XShm extension */
  if (!XShmQueryExtension (ptc_display))
    {
      XCloseDisplay (ptc_display);
      return PTC_FAILURE;
    }
#endif /* __PTC_XSHM__ */
  /* Get screen dimensions */
  ptc_screen_width = DisplayWidth (ptc_display, ptc_screen);
  ptc_screen_height = DisplayHeight (ptc_display, ptc_screen);

  /* Get the default root window */
  ptc_root_window = DefaultRootWindow (ptc_display);
  /* Initialize window's attribute structure */
  ptc_window_attributes.border_pixel = BlackPixel (ptc_display, ptc_screen);
  ptc_window_attributes.background_pixel =
    BlackPixel (ptc_display, ptc_screen);
  ptc_window_attributes.backing_store = NotUseful;

#ifdef __PTC_CENTER_WINDOW__
  /* Center the window on the screen */
  ptc_x_position = (ptc_screen_width - width) / 2;
  ptc_y_position = (ptc_screen_height - height) / 2;
#else
  /* Dock the window on the top-left corner */
  ptc_x_position = 0;
  ptc_y_position = 0;
#endif /* __PTC_CENTER_WINDOW__ */
  /* Create the window */
  ptc_window =
    XCreateWindow (ptc_display, ptc_root_window, ptc_x_position,
		   ptc_y_position, width, height, 0, ptc_depth, InputOutput,
		   ptc_visual, CWBackPixel | CWBorderPixel | CWBackingStore,
		   &ptc_window_attributes);
  /* Set the window's name */
  XStoreName (ptc_display, ptc_window, title);
  /* Tell the server to report only keypress-related events */
  XSelectInput (ptc_display, ptc_window, KeyPressMask | KeyReleaseMask);
    
  /* Initialize window's sizehint definition structure */
  ptc_window_sizehints.flags = PPosition | PMinSize | PMaxSize;
  ptc_window_sizehints.x = 0;
  ptc_window_sizehints.y = 0;
  ptc_window_sizehints.min_width = width;
  ptc_window_sizehints.max_width = width;
  ptc_window_sizehints.min_height = height;
  ptc_window_sizehints.max_height = height;
  /* Set the window's sizehint */
  XSetWMNormalHints (ptc_display, ptc_window, &ptc_window_sizehints);

  /* Set fullscreen stuff */
  if(fullscreen) {
	int vidmode, tmx,tmy;
    make_fullscreen_window();
	vidmode = ptc_get_nearest_mode (width, height, &tmx,&tmy);
	if (vidmode != -1) {
	  ptc_set_xvidmode (vidmode);
	  /* Clear event queue */
//	  XFlush (ptc_display);
	  /* Get the root window */
//	  ptc_root_window = XRootWindow (ptc_display, ptc_screen);
	  /* Get exclusive keyboard access */
//	  XGrabKeyboard (ptc_display, ptc_root_window, False, GrabModeAsync,
//					 GrabModeAsync, CurrentTime);
	  /* Get exclusive mouse access */
//	  XGrabPointer (ptc_display, ptc_root_window, True,
//					PointerMotionMask | ButtonPressMask | ButtonReleaseMask,
//					GrabModeAsync, GrabModeAsync, None, None, CurrentTime);
	}
  }
  /* Clear the window */
  XClearWindow (ptc_display, ptc_window);
  /* Put the window on top of the others */
  XMapRaised (ptc_display, ptc_window);
  /* Clear event queue */
  XFlush (ptc_display);
  /* Get the default graphic context */
  ptc_window_gc = DefaultGC (ptc_display, ptc_screen);
#ifdef __PTC_XSHM__
  /* Create an XShmImage */
  ptc_ximage =
    XShmCreateImage (ptc_display, ptc_visual, ptc_depth, ZPixmap, 0,
		     &ptc_shm_segment, width, height);
  /* Get a shared segment */
  ptc_shm_segment.shmid =
    shmget (IPC_PRIVATE, ptc_ximage->bytes_per_line * ptc_ximage->height,
	    IPC_CREAT | 0777);
  /* Initialize XShmImage data buffer pointer */
  ptc_ximage->data = shmat (ptc_shm_segment.shmid, 0, 0);
  /* Save buffer address */
  ptc_shm_segment.shmaddr = ptc_ximage->data;
  /* Put the segment in read/write */
  ptc_shm_segment.readOnly = False;
  /* Attach the segment to the display */
  if (!XShmAttach (ptc_display, &ptc_shm_segment))
    {
      /* Destroy the image */
      XDestroyImage (ptc_ximage);
      /* Detach the buffer from the segment */
      shmdt (ptc_shm_segment.shmaddr);
      /* Remove the segment */
      shmctl (ptc_shm_segment.shmid, IPC_RMID, 0);
      /* Destroy the window */
      XDestroyWindow (ptc_display, ptc_window);
      /* Close the display */
      XCloseDisplay (ptc_display);
      return PTC_FAILURE;
    }
#else
  /* Create an XImage */
  ptc_ximage =
    XCreateImage (ptc_display, CopyFromParent, ptc_depth, ZPixmap, 0, NULL,
		  width, height, 32, width * 2);
#endif /* __PTC_XSHM__ */
  /* Save windowsize values */
  ptc_viewport_width = width;
  ptc_viewport_height = height;
#endif /* __PTC_WINDOWED__ */

  ptc_opened = 1;
  return PTC_SUCCESS;
}

/* Update the screen */

int
ptc_update (void *buffer)
{
  if (ptc_opened) {
	if (fullscreen_cursor != None)
	  XWarpPointer(ptc_display, None, ptc_window,
				   0, 0, ptc_screen_width, ptc_screen_height,
				   1,1);
#ifdef __PTC_WINDOWED__
#ifdef __PTC_XSHM__
  /* Copy buffer data into the XShmImage */
#ifdef __PTC_ENABLE_CONVERSIONS__
  ptc_source_index = 0;
  ptc_destination_index = 0;
  /* Convert the image line by line */
  for (ptc_blitcounter = 0; ptc_blitcounter < ptc_viewport_height;
       ptc_blitcounter++)
    {
      /* Conversion */
      ptc_convert (buffer + ptc_source_index,
		   (ptc_ximage->data) + ptc_destination_index,
		   ptc_viewport_width);
      /* Pointers update */
      ptc_source_index += ptc_viewport_width * sizeof (int);
      ptc_destination_index += ptc_viewport_width * ptc_output_pitch;
    }
#else
  /* Blit the image */
  memcpy (ptc_ximage->data, buffer,
	  ptc_viewport_width * ptc_viewport_height * sizeof (int));
#endif /* __PTC_ENABLE_CONVERSIONS__ */
  /* Synchronize the event queue */
  XSync (ptc_display, 0);
  /* Put the buffer on the window */
  XShmPutImage (ptc_display, ptc_window, ptc_window_gc, ptc_ximage, 0, 0, 0,
		0, ptc_viewport_width, ptc_viewport_height, False);
#else
#ifdef __PTC_ENABLE_CONVERSIONS__
  ptc_ximage->data = ptc_buffer;
  ptc_source_index = 0;
  ptc_destination_index = 0;
  /* Convert the image line by line */
  for (ptc_blitcounter = 0; ptc_blitcounter < ptc_viewport_height;
       ptc_blitcounter++)
    {
      /* Conversion */
      ptc_convert (buffer + ptc_source_index,
		   (ptc_ximage->data) + ptc_destination_index,
		   ptc_viewport_width);
      /* Pointers update */
      ptc_source_index += ptc_viewport_width * sizeof (int);
      ptc_destination_index += ptc_viewport_width * ptc_output_pitch;
    }
#else
  /* Set XImage's data buffer value with the supplied buffer pointer */
  ptc_ximage->data = (char *) buffer;
#endif /* __PTC_ENABLE_CONVERSIONS__ */
  /* Put the buffer on the window */
  XPutImage (ptc_display, ptc_window, ptc_window_gc, ptc_ximage, 0, 0, 0, 0,
	     ptc_viewport_width, ptc_viewport_height);
  /* Check for incoming events */
  XFlush (ptc_display);
#endif /* __PTC_XSHM__ */
#endif /* __PTC_WINDOWED__ */
#ifdef __PTC_DGA__
  ptc_framebuffer_index = 0;
#ifdef __PTC_ENABLE_CONVERSIONS__
  ptc_source_index = 0;
  ptc_destination_index = 0;
#endif /* __PTC_ENABLE_CONVERSIONS__ */
  /* Do the blit line by line */
  for (ptc_blitcounter = 0; ptc_blitcounter < ptc_viewport_height;
       ptc_blitcounter++)
    {
#ifdef __PTC_ENABLE_CONVERSIONS__
      /* Conversion */
      ptc_convert (buffer + ptc_source_index,
		   ptc_framebuffer_address + ptc_framebuffer_start +
		   ptc_destination_index, ptc_viewport_width);
      /* Pointers update */
      ptc_source_index += ptc_viewport_width * sizeof (int);
      ptc_destination_index += ptc_framebuffer_pitch;
#else
      /* Copy the data */
      memcpy (ptc_framebuffer_address + ptc_framebuffer_start +
	      ptc_framebuffer_index, buffer,
	      ptc_viewport_width * sizeof (int));
      buffer += ptc_viewport_width * sizeof (int);
      /* Set up offsets */
      ptc_framebuffer_index += ptc_framebuffer_pitch;
#endif /* __PTC_ENABLE_CONVERSIONS__ */
    }
#endif /* __PTC_DGA__ */
  /* Process incoming events */
  if (ptc_process_events ())
    {
#ifdef __PTC_CLEANUP_CALLBACK__
      ptc_cleanup_callback ();
#endif /* __PTC_CLEANUP_CALLBACK__ */
      //ptc_close ();
      //exit (0);
    }
  }
  return PTC_SUCCESS;
}

/* Process events */

static int window_is_on_top = 1;

int
ptc_process_events (void)
{
  XEvent ptc_xevent;
  KeySym ptc_keysym;
  char string_buffer[4];
  int num_chars;
  /* Check if there are events waiting in the display's queue */
  if (XPending (ptc_display))
    {
      /* Get the next event in queue */
      XNextEvent (ptc_display, &ptc_xevent);
      /* Check if it's a keypress event */
      if (ptc_xevent.type == KeyPress)
	{
	  /* Get the keysym */

        num_chars = XLookupString(&ptc_xevent.xkey, string_buffer,
                                  4, &ptc_keysym, 0);
        
	  /* Check if the key pressed was a function one */
	  if ((ptc_keysym >> 8) == __PTC_FUNCTION_KEY__)
	    {
	      /* Check if it was the escape key */
	      if ((ptc_keysym & 0xFF) == __PTC_ESCAPE_KEY__) {
			escape_pressed = 1;
			return PTC_SUCCESS;
		  }
		  else if ((ptc_keysym & 0xFF) == __PTC_F1_KEY__) {
			f1_pressed = 1;
			return PTC_SUCCESS;
		  }
		  else if ((ptc_keysym & 0xFF) == __PTC_F2_KEY__) {
			f2_pressed = 1;
			return PTC_SUCCESS;
		  }
		  else if ((ptc_keysym & 0xFF) == __PTC_F3_KEY__) {
			f3_pressed = 1;
			return PTC_SUCCESS;
		  }
		  else if ((ptc_keysym & 0xFF) == __PTC_F4_KEY__) {
			f4_pressed = 1;
			return PTC_SUCCESS;
		  }
		  else if ((ptc_keysym & 0xFF) == __PTC_F5_KEY__) {
			f5_pressed = 1;
			return PTC_SUCCESS;
		  }
		  else if ((ptc_keysym & 0xFF) == __PTC_F6_KEY__) {
			f6_pressed = 1;
			return PTC_SUCCESS;
		  }
		  else if ((ptc_keysym & 0xFF) == __PTC_F7_KEY__) {
			f7_pressed = 1;
			return PTC_SUCCESS;
		  }
		  else if ((ptc_keysym & 0xFF) == __PTC_F8_KEY__) {
			f8_pressed = 1;
			return PTC_SUCCESS;
		  }
	    }
	  else {
		key_pressed [ptc_keysym & 0xFF] = 1;
		return PTC_SUCCESS;
	  }
/*	  if(num_chars == 1)
		{
		  if(string_buffer[0] == 'F')
			{
              if(window_is_on_top)
                XLowerWindow(ptc_display, ptc_window);
              else
                XRaiseWindow(ptc_display, ptc_window);
              window_is_on_top = !window_is_on_top;
              return PTC_SUCCESS;
			}
		}
*/
	}
      
    }
  return PTC_FAILURE;
}

/* Close the screen */

void
ptc_close (void)
{
  if (ptc_opened) {
	escape_pressed = 0;
#ifdef __PTC_WINDOWED__
#ifdef __PTC_XSHM__
	/* Detach the segment from the display */
	XShmDetach (ptc_display, &ptc_shm_segment);
	/* Destroy the XShmImage */
	XDestroyImage (ptc_ximage);
	/* Detach the buffer from the segment */
	shmdt (ptc_shm_segment.shmaddr);
	/* Remove the segment */
	shmctl (ptc_shm_segment.shmid, IPC_RMID, 0);
#else
	/* Restore XImage's buffer pointer */
	ptc_ximage->data = NULL;
	/* Destroy the XImage */
	XDestroyImage (ptc_ximage);
#endif /* __PTC_XSHM__ */
  /* Close the window */
  XDestroyWindow (ptc_display, ptc_window);
#endif /* __PTC_WINDOWED__ */
#ifdef __PTC_DGA__
  /* Exit from DGA mode */
  XF86DGADirectVideo (ptc_display, ptc_screen, False);
  /* Switch back to the previous video mode */
  XF86VidModeSwitchToMode (ptc_display, ptc_screen,
			   ptc_modeinfodata[ptc_previousmode_number]);
  /* Free modeinfo data */
  XFree (ptc_modeinfodata);
  /* Leave exclusive keyboard access */
  XUngrabKeyboard (ptc_display, CurrentTime);
  /* Leave exclusive mouse access */
  XUngrabPointer (ptc_display, CurrentTime);
#endif /* __PTC_DGA__ */

  /* Free fullscreen stuff */
  
  if(fullscreen_cursor != None)
    {
    XFreeCursor(ptc_display, fullscreen_cursor);
    fullscreen_cursor = None;
    }

  if(fullscreen_cursor_pixmap != None)
    {
    XFreePixmap(ptc_display, fullscreen_cursor_pixmap);
    fullscreen_cursor_pixmap = None;
    }
  if(fullscreen_cursor_mask != None)
    {
    XFreePixmap(ptc_display, fullscreen_cursor_mask);
    fullscreen_cursor_mask = None;
    }
  
  if (ptc_previousmode_number != -1) {
	ptc_restore_xvidmode ();
	ptc_previousmode_number = -1;
	/* Leave exclusive keyboard access */
//	XUngrabKeyboard (ptc_display, CurrentTime);
	/* Leave exclusive mouse access */
//	XUngrabPointer (ptc_display, CurrentTime);
  }
  
  /* Close the display */
  XCloseDisplay (ptc_display);
  /* Deallocate the buffer */
  if (ptc_buffer) {
	free (ptc_buffer);
  }
  }
  ptc_opened = 0;
}
